/************************************************************
* OpenGL light casting program                              *
* Rotate two light sources around a torus in 3D space,      *
* change the background color, the torus color, and view    *
* the torus from front, and perspective angles.             *  
* Full GLUI control panel provided for UI/UX functionality. *        
* Written by Chad Jordan                                    *
* Updated 12-18-2008                                        *
************************************************************/

#include <stdlib.h>
#include <cstdlib>
#include <stdio.h>
#include <GL/glui.h>
#include <GL/glut.h>

static int spinX, spinY = 0;  
static int spinX1, spinY1 = 0;
const float rotation_speed = 1.0f; 
int bg_selection = 0;
int torus_selection = 0;
float bgColor[3] = {0.0f, 0.0f, 0.0f}; 
float torusColor[3] = {1.0f, 0.0f, 0.0f}; 
float light0_color[3] = {1.0f, 1.0f, 1.0f}; 
float light1_color[3] = {1.0f, 1.0f, 1.0f};
float light0_intensity = 1.0f;
float light1_intensity = 1.0f; 
int wireframe = 0;
int light0_enabled = 1;
int light1_enabled = 0; 
int mainWindow;
int perspWindow;

float view_rotate[16] = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};
float bgColors[3][3] = {
    {0.0f, 0.0f, 0.0f},  
    {0.0f, 0.0f, 0.3f},   
    {0.5f, 0.5f, 0.5f}   
};

GLfloat torusColors[3][3] = {
    {1.0f, 0.0f, 0.0f},  
    {0.0f, 1.0f, 0.0f},   
    {0.0f, 0.0f, 1.0f}   
};

//Callback for GLUI controls
void control_cb(int control) {
    if (light0_enabled) {
        glutSetWindow(mainWindow);
        glEnable(GL_LIGHT0);
        glutSetWindow(perspWindow);
        glEnable(GL_LIGHT0);
    } else {
        glutSetWindow(mainWindow);
        glDisable(GL_LIGHT0);
        glutSetWindow(perspWindow);
        glDisable(GL_LIGHT0);
    }

    if (light1_enabled) {
        glutSetWindow(mainWindow);
        glEnable(GL_LIGHT1);
        glutSetWindow(perspWindow);
        glEnable(GL_LIGHT1);
    } else {
        glutSetWindow(mainWindow);
        glDisable(GL_LIGHT1);
        glutSetWindow(perspWindow);
        glDisable(GL_LIGHT1);
    }

    glutSetWindow(mainWindow);
    glutPostRedisplay();
    glutSetWindow(perspWindow);
    glutPostRedisplay();
}

void change_bg_color(int selection) {
     int mapped_selection;
     switch(selection) {
        case 3: mapped_selection = 0; break; 
        case 4: mapped_selection = 1; break; 
        case 5: mapped_selection = 2; break;  
        default: mapped_selection = 0;       
    }
    
    bgColor[0] = bgColors[bg_selection][0];
    bgColor[1] = bgColors[bg_selection][1];
    bgColor[2] = bgColors[bg_selection][2];
    
    glutPostRedisplay();
}

void change_torus_color(int selection) {
     int mapped_selection;
     switch(selection) {
        case 3: mapped_selection = 0; break; 
        case 4: mapped_selection = 1; break; 
        case 5: mapped_selection = 2; break;  
        default: mapped_selection = 0;       
    }
    
    torusColor[0] = torusColors[torus_selection][0];
    torusColor[1] = torusColors[torus_selection][1];
    torusColor[2] = torusColors[torus_selection][2];
    
    glutPostRedisplay();
}

void update_light_colors(int light_num) {
    if (light_num == 0) {
        GLfloat light0_diffuse[] = {light0_color[0], light0_color[1], light0_color[2], 1.0f};
        glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    } else {
        GLfloat light1_diffuse[] = {light1_color[0], light1_color[1], light1_color[2], 1.0f};
        glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
    }
    glutPostRedisplay();
}

void toggle_wireframe(int state) {
    wireframe = state; 
    glutPostRedisplay(); 
}

void init_glui() {
    GLUI *glui = GLUI_Master.create_glui("Control Panel", 0, 50, 100);
    glui->set_main_gfx_window(glutGetWindow());
    
    GLUI_Panel *info_panel = new GLUI_Panel(glui, "Program Info");
    new GLUI_StaticText(info_panel, "Welcome to Chad's Light Casting Program.                      ");
    new GLUI_StaticText(info_panel, "");
    new GLUI_StaticText(info_panel, "This program demonstrates the use of an OpenGL lighting model."); 
    new GLUI_StaticText(info_panel, "A torus is drawn with multiple light sources casting diffuse," );
    new GLUI_StaticText(info_panel, "ambient, and specular lighting effects to generate a material" );
    new GLUI_StaticText(info_panel, "characteristic.");
    new GLUI_StaticText(info_panel, "");
    new GLUI_StaticText(info_panel, "A Key light source illuminates the object by default, and a"   );
    new GLUI_StaticText(info_panel, "second Rim light source can be enabled behind the torus."      );
    new GLUI_StaticText(info_panel, "The user can rotate both light sources 180 degrees in positive");
    new GLUI_StaticText(info_panel, "or negative X and Y directions."                               );
    new GLUI_StaticText(info_panel, "");
    new GLUI_StaticText(info_panel, "Additional controls are also available below."                        );
        
    //Horizontal panel to hold both color panels
    GLUI_Panel *color_container = new GLUI_Panel(glui, "Color Controls");
    color_container->set_alignment(GLUI_ALIGN_CENTER);
    
    GLUI_Panel *bg_panel = new GLUI_Panel(color_container, "Background Color");
    bg_panel->set_alignment(GLUI_ALIGN_LEFT);
    GLUI_RadioGroup *colorGroup = new GLUI_RadioGroup(bg_panel, &bg_selection, 3, change_bg_color);
    new GLUI_RadioButton(colorGroup, "Black");
    new GLUI_RadioButton(colorGroup, "Dark Blue");
    new GLUI_RadioButton(colorGroup, "Gray");
    
    GLUI_Column *column = new GLUI_Column(color_container, true);
    
    GLUI_Panel *torus_panel = new GLUI_Panel(color_container, "Torus Color");
    torus_panel->set_alignment(GLUI_ALIGN_RIGHT);
    GLUI_RadioGroup *torusGroup = new GLUI_RadioGroup(torus_panel, &torus_selection, 3, change_torus_color);
    new GLUI_RadioButton(torusGroup, "Red");
    new GLUI_RadioButton(torusGroup, "Green");
    new GLUI_RadioButton(torusGroup, "Blue"); 
    
    GLUI_Panel *lights_panel = new GLUI_Panel(glui, "Light Controls");

    GLUI_Panel *light0_panel = new GLUI_Panel(lights_panel, "Key Light");
    GLUI_Checkbox *light0_toggle = new GLUI_Checkbox(light0_panel, "Enable Key Light", &light0_enabled);
    GLUI_Spinner *light0_r = new GLUI_Spinner(light0_panel, "Red:", &light0_color[0]);
    GLUI_Spinner *light0_g = new GLUI_Spinner(light0_panel, "Green:", &light0_color[1]);
    GLUI_Spinner *light0_b = new GLUI_Spinner(light0_panel, "Blue:", &light0_color[2]);
    GLUI_Spinner *intensity0_spinner = new GLUI_Spinner(light0_panel, "Intensity:", &light0_intensity);
    intensity0_spinner->set_float_limits(0.0f, 2.0f);
    intensity0_spinner->set_speed(0.1);
    light0_r->set_float_limits(0.0f, 1.0f);
    light0_g->set_float_limits(0.0f, 1.0f);
    light0_b->set_float_limits(0.0f, 1.0f);

    new GLUI_Column(lights_panel, false);

    GLUI_Panel *light1_panel = new GLUI_Panel(lights_panel, "Rim Light");
    GLUI_Checkbox *light1_toggle = new GLUI_Checkbox(light1_panel, "Enable Rim Light", &light1_enabled);
    GLUI_Spinner *light1_r = new GLUI_Spinner(light1_panel, "Red:", &light1_color[0]);
    GLUI_Spinner *light1_g = new GLUI_Spinner(light1_panel, "Green:", &light1_color[1]);
    GLUI_Spinner *light1_b = new GLUI_Spinner(light1_panel, "Blue:", &light1_color[2]);
    GLUI_Spinner *intensity1_spinner = new GLUI_Spinner(light1_panel, "Intensity:", &light1_intensity);
    intensity1_spinner->set_float_limits(0.0f, 2.0f);
    intensity1_spinner->set_speed(0.1);
    light1_r->set_float_limits(0.0f, 1.0f);
    light1_g->set_float_limits(0.0f, 1.0f);
    light1_b->set_float_limits(0.0f, 1.0f);
    
    GLUI_Panel *rotations_container = new GLUI_Panel(glui, "Rotation Controls");
    GLUI_Panel *rotation_panel = new GLUI_Panel(rotations_container, "Key Light");
    GLUI_Spinner *x_spinner = new GLUI_Spinner(rotation_panel, "Rotate X:", &spinX, 1, control_cb);
    x_spinner->set_float_limits(-180.0, 180.0, GLUI_LIMIT_WRAP);
    x_spinner->set_speed(0.3);

    GLUI_Spinner *y_spinner = new GLUI_Spinner(rotation_panel, "Rotate Y:", &spinY, 2, control_cb);
    y_spinner->set_float_limits(-180.0, 180.0, GLUI_LIMIT_WRAP);
    y_spinner->set_speed(0.3);

    new GLUI_Column(rotations_container, false);

    GLUI_Panel *rotation_panel1 = new GLUI_Panel(rotations_container, "Rim Light");
    GLUI_Spinner *x1_spinner = new GLUI_Spinner(rotation_panel1, "Rotate X:", &spinX1, 1, control_cb);
    x1_spinner->set_float_limits(-180.0, 180.0, GLUI_LIMIT_WRAP);
    x1_spinner->set_speed(0.3);

    GLUI_Spinner *y1_spinner = new GLUI_Spinner(rotation_panel1, "Rotate Y:", &spinY1, 2, control_cb);
    y1_spinner->set_float_limits(-180.0, 180.0, GLUI_LIMIT_WRAP);
    y1_spinner->set_speed(0.3);
    
    GLUI_Panel *view_panel = new GLUI_Panel(glui, "Perspective Control");
    GLUI_Rotation *view_rot = new GLUI_Rotation(view_panel, "Rotate", view_rotate);
    
    view_rot->set_spin(1.0);
    new GLUI_Checkbox(glui, "Toggle Wireframe", &wireframe);
    new GLUI_Button(glui, "Exit", 0, (GLUI_Update_CB)exit);
}

//Initialize material property, light source, lighting model, and depth buffer.
void init(void) {
   GLfloat mat_specular[] = {1.0, 1.0, 1.0, 1.0};
   GLfloat mat_shininess[] = {50.0};
   GLfloat light_position[] = {1.0, 1.0, 1.0, 0.0};
   GLfloat light1_position[] = { -1.0, 1.0, 1.0, 0.0 };
   GLfloat mat_diffuse[] = {.5, 1.0, .5, 0.0};

   glShadeModel (GL_SMOOTH);

   glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
   glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
   glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);

   glLightfv(GL_LIGHT0, GL_POSITION, light_position);
   glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHT1);
   glEnable(GL_DEPTH_TEST);
   glMatrixMode(GL_PROJECTION);
   
}

void displayPersp(void) {
     if (light0_enabled) {
        glEnable(GL_LIGHT0);
    } else {
        glDisable(GL_LIGHT0);
    }
    
    if (light1_enabled) {
        glEnable(GL_LIGHT1);
    } else {
        glDisable(GL_LIGHT1);
    }
    GLfloat position[] = { 0.0, 0.0, 1.5, 1.0 };
    GLfloat position1[] = {0.0f, 2.0f, -2.0f, 1.0f};
    
    glClearColor(bgColor[0], bgColor[1], bgColor[2], 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glPushMatrix();
    //Perspective camera position
    gluLookAt(5.0, 5.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    glMultMatrixf(view_rotate); 
    
    //Key Light (Light 0)
    glPushMatrix();
    glRotated((GLdouble) spinY, 1.0, 0.0, 0.0);
    glRotated((GLdouble) spinX, 0.0, 1.0, 0.0);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
    glTranslated(0.0, 0.0, 1.5);
    glDisable(GL_LIGHTING);
    glColor3f(0.0, 1.0, 1.0);
    glutWireCube(0.1);
    glEnable(GL_LIGHTING);
    glPopMatrix();
    
    //Rim Light (Light 1)
    glPushMatrix();
    glRotated((GLdouble) spinY1, 1.0, 0.0, 0.0);
    glRotated((GLdouble) spinX1, 0.0, 1.0, 0.0);
    glLightfv(GL_LIGHT1, GL_POSITION, position1);
    glTranslated(0.0, 0.0, -1.5);
    glDisable(GL_LIGHTING);
    glColor3f(1.0, 1.0, 0.0);
    glutWireCube(0.1);
    glEnable(GL_LIGHTING);
    glPopMatrix();
    
    GLfloat light0_diffuse[] = {1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f};
    GLfloat light0_specular[] = {1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);

    //Increase the intensity of Light 1 (Rim Light)
    GLfloat rim_diffuse[] = {2.0f * light1_intensity, 2.0f * light1_intensity, 3.0f * light1_intensity, 1.0f};
    GLfloat rim_specular[] = {3.0f * light1_intensity, 3.0f * light1_intensity, 3.0f * light1_intensity, 1.0f};
    GLfloat rim_ambient[] = {0.2f * light1_intensity, 0.2f * light1_intensity, 0.4f * light1_intensity, 1.0f};
    glLightfv(GL_LIGHT1, GL_DIFFUSE, rim_diffuse);
    glLightfv(GL_LIGHT1, GL_SPECULAR, rim_specular);
    glLightfv(GL_LIGHT1, GL_AMBIENT, rim_ambient);
    
    //Rim light position and direction
    GLfloat spot_direction[] = {0.0f, -0.8f, 1.0f};
    glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 30.0);
    glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 20.0);
    glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, spot_direction);
    
    //Material properties
    GLfloat mat_ambient[] = {0.2f * light0_intensity, 0.2f * light0_intensity, 0.2f * light0_intensity, 1.0f};
    GLfloat mat_diffuse[] = {torusColor[0] * light0_intensity, torusColor[1] * light0_intensity, torusColor[2] * light0_intensity, 1.0f};
    GLfloat mat_specular[] = {1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f};
    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    
    //Draw the torus
    if (wireframe) {
        glutWireTorus(0.275, 0.85, 85, 85);
    } else {
        glutSolidTorus(0.275, 0.85, 85, 85);
    }
    
    glPopMatrix();
    glutSwapBuffers();
}

void display(void) {
   GLfloat position[] = { 0.0, 0.0, 1.5, 1.0 };
   GLfloat position1[] = {-2.5f, 1.0f, -2.0f, 1.0f};
   glClearColor(bgColor[0], bgColor[1], bgColor[2], 1.0f);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glPushMatrix ();
   gluLookAt (0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
   
   GLfloat light0_diffuse[] = {1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f};
   GLfloat light0_specular[] = {1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f};
   GLfloat light0_ambient[] = {0.2f * light0_intensity, 0.2f * light0_intensity, 0.2f * light0_intensity, 1.0f};
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);

   glPushMatrix ();
   glRotated ((GLdouble) spinY, 1.0, 0.0, 0.0);
   glLightfv (GL_LIGHT0, GL_POSITION, position);

   glRotated ((GLdouble) spinX, 0.0, 1.0, 0.0);
   glLightfv (GL_LIGHT0, GL_POSITION, position);
   glTranslated (0.0, 0.0, 1.5);
   
   glDisable (GL_LIGHTING);
   glColor3f (0.0, 1.0, 1.0);
   glutWireCube (0.1);
   glEnable (GL_LIGHTING);
   glPopMatrix ();
   
   if (light0_enabled) {
        glEnable(GL_LIGHT0);
   } else {
        glDisable(GL_LIGHT0);
   }
   
   glPushMatrix();
   GLfloat light1_diffuse[] = {0.6f * light1_intensity, 0.6f * light1_intensity, 1.0f * light1_intensity, 1.0f};
   GLfloat light1_ambient[] = {0.1f * light1_intensity, 0.1f * light1_intensity, 0.2f * light1_intensity, 1.0f};
   GLfloat light1_specular[] = {1.0f * light1_intensity, 1.0f * light1_intensity, 1.0f * light1_intensity, 1.0f}; 
   glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
   glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);
   glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient);
   
   glRotated((GLdouble) spinY1, 1.0, 0.0, 0.0);
   glRotated((GLdouble) spinX1, 0.0, 1.0, 0.0);
   glLightfv(GL_LIGHT1, GL_POSITION, position1);
   
   light1_color[0] = 0.6f;
   light1_color[1] = 0.6f;
   light1_color[2] = 1.0f;
   
   GLfloat spot_direction[] = {0.8f, -0.2f, 0.8f}; 
   glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 60.0);        
   glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 5.0);  
   glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, spot_direction);
   
   //Rim Light1 position behind torus
   glDisable(GL_LIGHTING);
   glColor3f(1.0, 1.0, 0.0);
   glTranslated(0.0, 0.0, -1.5);
   glutWireCube(0.1);
   glEnable(GL_LIGHTING);
   glPopMatrix();
   
   if (light1_enabled) {
        glEnable(GL_LIGHT1);
   } else {
        glDisable(GL_LIGHT1);
   }
    
    update_light_colors(0);
    update_light_colors(1);
    
    GLfloat mat_ambient[] = {0.2f * light0_intensity, 0.2f * light0_intensity, 0.2f * light0_intensity, 1.0f};
    GLfloat mat_diffuse[] = {torusColor[0] * light0_intensity, torusColor[1] * light0_intensity, torusColor[2] * light0_intensity, 1.0f};
    GLfloat mat_specular[] = {1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f * light0_intensity, 1.0f};

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    
   if (wireframe) {
        glutWireTorus(0.275, 0.85, 85, 85); 
    } else {
        glutSolidTorus (0.275, 0.85, 85, 85);
    }
      
   glPopMatrix ();
   glFlush ();
   glutSwapBuffers();
}

void reshape (int w, int h) {
   glViewport (0, 0, (GLsizei) w, (GLsizei) h);
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity();
   if (w <= h)
      glOrtho (-1.5, 1.5, -1.5*(GLfloat)h/(GLfloat)w,
         1.5*(GLfloat)h/(GLfloat)w, -10.0, 10.0);
   else
      glOrtho (-1.5*(GLfloat)w/(GLfloat)h,
         1.5*(GLfloat)w/(GLfloat)h, -1.5, 1.5, -10.0, 10.0);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

void keyboard(unsigned char key, int x, int y) {
    switch (key) {
      case 'X':
         spinX = (spinX + 5) % 360;
         glutPostRedisplay();
         default:
         break;  
      case 'x':
         spinX = (spinX - 5) % 360;
         glutPostRedisplay();
         break; 
      case 'Y':
         spinY = (spinY + 5) % 360;
         glutPostRedisplay();
         break;  
      case 'y':
         spinY = (spinY - 5) % 360;
         glutPostRedisplay();
         break; 
   }
}

int main(int argc, char** argv) {
   glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(350, 350);
    glutInitWindowPosition(500, 100);
    mainWindow = glutCreateWindow("Front View");
    init();
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);

    glutInitWindowSize(350, 350);
    glutInitWindowPosition(500, 500);
    perspWindow = glutCreateWindow("Perspective View");
    init();
    glutDisplayFunc(displayPersp);
    glutReshapeFunc(reshape);

    init_glui();
    glutMainLoop();
    return 0;
}
